"""
Useful Pel tasks.
"""
__all__ = [
    "Shell",
]

import subprocess
from typing import Any, List, NamedTuple, Union

from pel.core import Task
from pel.filesystem import any_source_newer_or_eq_than_target
from pel.types import enforce_list_of_str


class ShellResult(NamedTuple):
    """The result of a single shell command."""

    cmd: str
    proc: "Union[subprocess.CompletedProcess[bytes], subprocess.CompletedProcess[str]]"


class Shell(Task):
    """Run a shell command."""

    cmd: Union[str, List[str]]
    src: Union[str, List[str]] = ""
    target: Union[str, List[str]] = ""
    text: bool = False
    quiet: bool = False
    check: bool = False

    @classmethod
    def is_expired(cls) -> bool:
        if cls.src and cls.target:
            sources = enforce_list_of_str(cls.src)
            targets = enforce_list_of_str(cls.target)
            return any_source_newer_or_eq_than_target(sources=sources, targets=targets)
        return True

    @classmethod
    def run(cls) -> Any:
        cmds = enforce_list_of_str(cls.cmd)
        return [
            ShellResult(
                cmd=cmd,
                proc=subprocess.run(
                    cmds,
                    shell=True,
                    text=cls.text,
                    capture_output=cls.quiet,
                    check=cls.check,
                ),
            )
            for cmd in cmds
        ]



"""
Defines the Pel runner.
"""
import typing
from pathlib import Path

import pel.core
import pel.tasks
from pel.tasks import *


def run(
    *,
    graph: typing.Optional[pel.core.Graph] = None,
    args: typing.Optional[typing.List[str]] = None,
    filename: str = "build.py",
    encoding: str = "utf-8",
) -> int:
    """Runs a Pel build file."""
    if graph is None:
        graph = pel.core.Graph()
    filepath = str((Path.cwd() / filename).absolute())
    try:
        with open(filepath, encoding=encoding) as handle:
            source = handle.read()
    except FileNotFoundError:
        print("Could not find a Pel Build File at:")
        print("\t", filepath, sep="")
        print("\nNeed help writing a Build File?")
        print("You can find instructions at https://github.com/neocrym/pel\n")
        return 1
    exec_globals = {
        key: val for key, val in pel.tasks.__dict__.items() if key in pel.tasks.__all__
    }
    exec(
        compile(source=source, filename=filename, mode="exec"),
        exec_globals,
        exec_globals,
    )
    graph.add_tasks_from_scope(exec_globals)
    parser = pel.core.ArgParser(graph=graph)
    parser.interpret_args(args=args)
    return 0



"""Utilities for typing Python code."""
from typing import List, Union


def enforce_list_of_str(inp: Union[str, List[str]]) -> List[str]:
    """Always convert strings into lists of strings."""
    if isinstance(inp, str):
        return [inp]
    return inp



"""Utils for Docker and Docker Compose."""
import json
import shlex
import subprocess
from typing import Optional

from pel.datetime import str_datetime_to_unixtime


def docker_image_last_modified(
    *, image_name: str, docker_path: str = "docker"
) -> Optional[float]:
    """Return the Unix time of when a given Docker image was last modified."""
    quoted_docker_path = shlex.quote(docker_path)
    proc = subprocess.run(
        (quoted_docker_path, "inspect", image_name),
        text=True,
        capture_output=True,
        check=True,
    )
    if proc.returncode != 0:
        print(proc.stderr)
        return None
    metadata = json.loads(proc.stdout)
    return str_datetime_to_unixtime(metadata[0]["Created"])



#!/usr/bin/env python3
"""
The Pel command line entrypoint.
"""
import sys

import pel.run


def run() -> None:
    """Run the Pel CLI and exit with a status code."""
    sys.exit(pel.run.run())


if __name__ == "__main__":
    run()



"""Utilities for filesystem tasks."""
import os
import stat
from typing import Iterator, Optional, Sequence, Tuple


def path_iterator(root: str) -> Iterator[Tuple[str, float]]:
    """
    Iterates over the filesystem in top-down order, yielding
    tuples of ``(filenames, last_modified)`` times.

    The iterator includes the root path.
    """

    # pylint: disable=unused-argument
    def _path_iterator(
        path: str, _last_mod: Optional[float] = None
    ) -> Iterator[Tuple[str, float]]:
        """Internal iterator for yielding subdirectories."""
        with os.scandir(path) as scandir_iter:
            dir_paths = []
            for obj in scandir_iter:
                obj_stat = obj.stat(follow_symlinks=True)
                yield (obj.path, obj_stat.st_mtime)
                if obj.is_dir():
                    dir_paths.append(obj.path)
            for dir_path in dir_paths:
                yield from path_iterator(dir_path)

    root_abspath = os.path.abspath(root)
    root_stat = os.stat(root_abspath, follow_symlinks=True)
    yield (root_abspath, root_stat.st_mtime)
    if stat.S_ISDIR(os.stat(root_abspath).st_mode):
        yield from _path_iterator(root_abspath)


def filesystem_path_is_newer_or_eq(path: str, last_modified: float) -> bool:
    """
    Returns ``True`` if any file or directory in ``path``
    has a modification time >= ``last_modified``.

    This function also checks the ``path`` directory itself.
    """
    for _, path_last_mod in path_iterator(path):
        if path_last_mod >= last_modified:
            return True
    return False


def any_source_newer_or_eq_than_target(
    *, sources: Sequence[str], targets: Sequence[str]
) -> bool:
    """
    Returns ``True`` if any source path is newer than any target path.
    """
    newest_target_lm: Optional[float] = None
    try:
        for target in targets:
            for _, iter_lm in path_iterator(target):
                if newest_target_lm is None or newest_target_lm < iter_lm:
                    newest_target_lm = iter_lm
    except FileNotFoundError:
        return True
    if newest_target_lm is None:
        return True
    return any(
        filesystem_path_is_newer_or_eq(source, newest_target_lm) for source in sources
    )



"""Utils for dates and times."""
import datetime
import time

from pel._vendor.dateutil.parser import parse as dateutil_parse


def str_datetime_to_unixtime(input_str: str) -> float:
    """
    Parse a datetime from a string. Return a UTC Unix timestamp.
    """
    datetime_obj: datetime.datetime = dateutil_parse(input_str)  # type: ignore
    return time.mktime(datetime_obj.timetuple())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pel',
 'pel._vendor',
 'pel._vendor.dateutil',
 'pel._vendor.dateutil.parser',
 'pel._vendor.dateutil.tz',
 'pel._vendor.dateutil.zoneinfo',
 'pel._vendor.graphlib']

package_data = \
{'': ['*'],
 'pel._vendor': ['graphlib_backport-1.0.3.dist-info/*',
                 'python_dateutil-2.8.2.dist-info/*']}

entry_points = \
{'console_scripts': ['pel = pel.console:run']}

setup_kwargs = {
    'name': 'pel',
    'version': '0.0.1',
    'description': 'The most elegant build system.',
    'long_description': '# Pel is the most elegant build system\n## Introduction\n\nPel is an easy-to-use build system/task runner written in Python. It is designed to be more advanced than task runners like [Make](https://www.gnu.org/software/make/) or [Invoke](https://www.pyinvoke.org/), but much simpler than complex build systems like [CMake](https://cmake.org/) or [Bazel](https://bazel.build/).\n\n## Installation\n\n### Installation from pip\n\n```\npip3 install --user pel\n```\n\nIt is safe to install Pel as a global Python package. **Installing Pel will never modify any other Python packages on your computer.**\n\n## Development status\n\nPel is still in the early stages of development. It is not currently ready for production use.\n\n## Features\n\n### Cross-platform support\n\nPel is written in pure Python, and is intended to work on any operating system supported by Python, such as:\n* Linux\n* Windows\n* macOS\n* FreeBSD\n\n## Why we made Pel\n\nMost build systems are either **too simple** or **too complex**.\n\n * A **simple** build system, like [Make](https://www.gnu.org/software/make/) or [Invoke](https://www.pyinvoke.org/) makes it easy to run arbitrary shell commands, but makes it hard to add non-trivial dependency management and build caching\n * A **complex** build system, like [CMake](https://cmake.org/) or [Bazel](https://bazel.build/), offers sophisticated dependency management and build caching, but only for predefined types of build targets. These build systems are excellent choices for building a large C++ monorepo, but can be unwieldy to integrate with arbitrary commands and obscure build tools.\n\nPel is designed to be the happy medium between simple and complex.\n',
    'author': 'Neocrym Records Inc.',
    'author_email': 'engineering@neocrym.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/neocrym/pel',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""The Pel Graph, Task, and ArgParser models."""
import argparse
import collections
import threading
from enum import Enum, auto
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    OrderedDict,
    Sequence,
    Set,
    Type,
    Union,
    cast,
)

from pel._vendor.graphlib import TopologicalSorter  # type: ignore


def _get_recursive_need_names_from_tasks(end_tasks: Sequence[Type["Task"]]) -> Set[str]:
    def _recursive(tasks: Set[Type["Task"]]) -> Iterator[str]:
        for task in tasks:
            yield task.__name__
            yield from _recursive(set(task._get_needs()))

    return set(_recursive(set(end_tasks)))


class TaskExecuted(Enum):
    """Describes whether a Task has been executed."""

    NOT_RUN = auto()
    NOT_EXPIRED = auto()
    EXECUTED = auto()
    RAISED_EXCEPTION = auto()
    SKIPPED = auto()


class Task:
    """Subclass this to create your own Task types."""

    needs: Union[List[Type["Task"]], Type["Task"], None] = None

    _result: Any = NotImplemented
    _exception: Optional[BaseException] = None
    _executed: TaskExecuted = TaskExecuted.NOT_RUN

    @classmethod
    def _get_needs(cls) -> List[Type["Task"]]:
        """
        Getter method for this Task's dependencies.
        """
        if not cls.needs:
            return []
        if isinstance(cls.needs, Task.__class__):  # type: ignore
            return [cls.needs]  # type: ignore
        if not isinstance(cls.needs, list):
            cast(List[Type["Task"]], cls.needs)
            raise TypeError(f"bad needs type {cls.needs}")
        for task in cls.needs:
            if not isinstance(task, Task.__class__):  # type: ignore
                raise TypeError(f"bad needs task type {task}")
        return cls.needs

    @classmethod
    def is_expired(cls) -> bool:
        """
        Override this to return False when this Task does not need to run.

        By default, this classmethod always returns True, which is handy
        for when you always want a Task to run.
        """
        return True

    @classmethod
    def run(cls) -> None:
        """
        Override this with the actual business logic that you want this Task to run.
        """
        raise NotImplementedError("HI")


class Graph:
    """Keeps track of :py:class:`Task` classes and executes them."""

    def __init__(self) -> None:
        self._lock = threading.Lock()
        self.clear()

    def clear(self) -> None:
        """Clear all references in the graph."""
        self._registry: Dict[str, Type[Task]] = {}

    def get_task(self, name: str) -> Type[Task]:
        """Get a Task class by its name."""
        return self._registry[name]

    def add_task(self, cls: Type[Task]) -> None:
        """
        Add a Task class to the graph.

        Args:
            cls: The Task class. Do not pass an instance of the class.
        """
        with self._lock:
            for base in cls.__bases__:
                self._registry.pop(base.__name__, None)
            self._registry[cls.__name__] = cls

    def add_tasks_from_scope(self, scope: Mapping[str, Any]) -> None:
        """Search a variable scope dictionary for any Task subclass objects."""
        for obj in scope.values():
            if isinstance(obj, Task.__class__):  # type: ignore
                self.add_task(obj)

    def remove_task(self, name: str) -> None:
        """Remove a Task by name."""
        with self._lock:
            self._registry.pop(name, None)

    def get_all_tasks(self) -> Dict[str, Type[Task]]:
        """Return a dictionary mapping Task names to Task classes."""
        return self._registry

    def _get_recursive_need_names(
        self, end_task_names: Union[Sequence[str], str]
    ) -> Set["str"]:
        """List the names of all of the tasks needed to execute these task names."""
        if isinstance(end_task_names, str):
            end_task_names = [end_task_names]
        return _get_recursive_need_names_from_tasks(
            [self._registry[name] for name in end_task_names]
        )

    def _get_needs_map(
        self, end_task_names: Union[Sequence[str], str] = ""
    ) -> OrderedDict[str, Set[str]]:
        """Return a dict mapping task names to their dependencies' names."""
        needs_map: OrderedDict[str, Set[str]] = collections.OrderedDict()
        if end_task_names:
            tasks = self._get_recursive_need_names(end_task_names)
        else:
            tasks = set(self._registry.keys())
        for name in sorted(tasks):
            task = self._registry[name]
            needs_map[name] = {dep.__name__ for dep in task._get_needs()}
        return needs_map

    def get_needs_map(
        self, end_task_names: Union[Sequence[str], str] = ""
    ) -> Dict[str, Set[str]]:
        """
        Return a dict mapping task names to their dependencies' names.

        Acquires the lock.
        """
        with self._lock:
            return self._get_needs_map(end_task_names)

    def _get_sorted_task_names(
        self, end_task_names: Union[Sequence[str], str] = ""
    ) -> List[str]:
        """Returns all task names, topologically sorted by dependencies."""
        return list(
            TopologicalSorter(self._get_needs_map(end_task_names)).static_order()
        )

    def get_sorted_task_names(
        self, end_task_names: Union[Sequence[str], str] = ""
    ) -> List[str]:
        """
        Returns all task names, topologically sorted by dependencies.

        Acquires the lock.
        """
        with self._lock:
            return self._get_sorted_task_names(end_task_names)

    def _get_sorted_tasks(
        self, end_task_names: Union[Sequence[str], str] = ""
    ) -> List[Type[Task]]:
        """
        Returns all task class instances, topologically sorted by dependencies.
        """
        return [
            self._registry[name] for name in self._get_sorted_task_names(end_task_names)
        ]

    def get_sorted_tasks(
        self, end_task_names: Union[Sequence[str], str] = ""
    ) -> List[Type[Task]]:
        """
        Returns all task class instances, topologically sorted by dependencies.

        Acquires the lock.
        """
        with self._lock:
            return self._get_sorted_tasks(end_task_names)

    def _execute(self, end_task_names: Union[Sequence[str], str] = "") -> None:
        """
        Executes the task names provided, including dependencies.

        If no names are provided, we execute all of the tasks.
        """
        for task in self._get_sorted_tasks(end_task_names):
            if task.is_expired():
                print(task.__name__, "[running]")
                try:
                    task._result = task.run()
                except BaseException as exc:
                    task._executed = TaskExecuted.RAISED_EXCEPTION
                    task._exception = exc
                    raise
                task._executed = TaskExecuted.EXECUTED
            else:
                print(task.__name__, "[not expired]")
                task._executed = TaskExecuted.NOT_EXPIRED

    def execute(self, end_task_names: Union[Sequence[str], str] = "") -> None:
        """
        Executes the task names provided, including dependencies.

        If no names are provided, we execute all of the tasks.

        Acquires the lock.
        """
        with self._lock:
            self._execute(end_task_names)


class ArgParser:
    """Parses shell arguments to run Tasks."""

    def __init__(self, *, graph: Graph) -> None:
        """
        Initialize the argument parser.

        Args:
            graph: The Graph to parse.
        """
        self._graph: Graph = graph
        self._parser = argparse.ArgumentParser()
        group = self._parser.add_mutually_exclusive_group(required=True)
        group.add_argument("--list", "-l", action="store_true")
        group.add_argument("--all", "-a", action="store_true")
        group.add_argument("tasks", metavar="task", type=str, nargs="*", default=())

    def parse_args(self, *, args: Optional[List[str]] = None) -> argparse.Namespace:
        """Parse the args."""
        return self._parser.parse_args(args=args)

    def interpret_args(self, *, args: Optional[List[str]] = None) -> None:
        """Parse the args and do what they say."""
        parsed_args = self.parse_args(args=args)
        if parsed_args.list:
            self.list_tasks()
            return
        if parsed_args.all:
            self.execute_tasks()
            return
        if not parsed_args.tasks:
            self.print_help()
            return
        self.execute_tasks(parsed_args.tasks)
        return

    def list_tasks(self) -> None:
        """Print the Tasks in our Graph to standard output."""
        print("Detected tasks:\n")
        for name, task in sorted(self._graph.get_all_tasks().items()):
            dash_name = name.replace("_", "-")
            task_help = (task.__doc__ or "").strip().split("\n")[0].strip()
            if task_help:
                print(f"{dash_name}:\n    {task_help}\n")
            else:
                print(f"{dash_name}:\n")

    def print_help(self) -> None:
        """Print this argparser's Help message to standard output."""
        return self._parser.print_help()

    def execute_tasks(self, dash_names: Union[Sequence[str], str] = "") -> None:
        """Execute the tasks."""
        end_task_names = [name.replace("-", "_") for name in dash_names]
        self._graph.execute(end_task_names=end_task_names)
